package org.processmining.streamconformance.regions.plugin.cg;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.transitionsystem.TransitionSystemConnection;
import org.processmining.models.graphbased.directed.DirectedGraphElementWeights;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.transitionsystem.AcceptStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;
import org.processmining.models.graphbased.directed.transitionsystem.StartStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.streamconformance.regions.gui.panels.CostsSetup;

public class CoverabilityGraphExtender {

	private ExtendedCoverabilityGraph coverabilityGraph = null;
	private StartStateSet startIds = new StartStateSet();
	private AcceptStateSet acceptIds = new AcceptStateSet();
	private DirectedGraphElementWeights weights = new DirectedGraphElementWeights();
	private int costWrongFlow = 1;
	private int costActivityNotInProcess = 1;

	// @Plugin(
	// name = "Petri net to Extended Coverability graph",
	// parameterLabels = { "Model", "Marking" },
	// returnLabels = {
	// "Extended Coverability Graph",
	// "Transition weights",
	// "Set of starting states",
	// "Set of accepting states",
	// "Coverability Graph (without extension)"
	// },
	// returnTypes = {
	// ExtendedCoverabilityGraph.class,
	// DirectedGraphElementWeights.class,
	// StartStateSet.class,
	// AcceptStateSet.class,
	// CoverabilityGraph.class
	// })
	// @UITopiaVariant(
	// author = "A. Burattin",
	// affiliation = "",
	// email = "")
	public Object[] plugin(UIPluginContext context, Petrinet model, Marking state) throws ConnectionCannotBeObtained {

		// build coverability graph
		CoverabilityGraph originalGraph = context.tryToFindOrConstructFirstNamedObject(CoverabilityGraph.class,
				"Construct Coverability Graph of a Petri Net", null, null, model, state);
		ConnectionManager cm = context.getConnectionManager();
		TransitionSystemConnection tsc = cm.getFirstConnection(TransitionSystemConnection.class, context, originalGraph);

		StartStateSet originalStartIds = (StartStateSet) tsc.getObjectWithRole(TransitionSystemConnection.STARTIDS);
		AcceptStateSet originalAcceptIds = (AcceptStateSet) tsc.getObjectWithRole(TransitionSystemConnection.ACCEPTIDS);

		// dialog configuration
		CostsSetup cs = new CostsSetup();
		InteractionResult result = context.showWizard("Costs Setup", true, true, cs);

		if (result.equals(InteractionResult.FINISHED)) {
			// perform the actual extension
			Object[] objects = extend(originalGraph, originalStartIds, originalAcceptIds, cs.getCostWrongFlow(),
					cs.getCostActivityNotInProcess());
			// connect the various objects
			context.addConnection(new TransitionSystemConnection(coverabilityGraph, weights, startIds, acceptIds));
			// let the plugin just return the results
			return objects;
		}
		return null;
	}

	public Object[] extend(CoverabilityGraph originalGraph, StartStateSet originalStartIds, AcceptStateSet originalAcceptIds,
			int costWrongFlow,
			int costActivityNotInProcess) {

		this.costWrongFlow = costWrongFlow;
		this.costActivityNotInProcess = costActivityNotInProcess;

		// clone the provided graph in order to create the new
		cloneGraph(originalGraph, originalStartIds, originalAcceptIds);

		// bypass the silent transitions (in this phase we add transitions -> possible violation of determininsm)
		bypassSilent();

		// extend the graph in order to deal with deviations by adding weights
		addWeights();

		// prepare the final stuff to be returned
		Object[] objects = new Object[5];
		objects[0] = coverabilityGraph;
		objects[1] = weights;
		objects[2] = startIds;
		objects[3] = acceptIds;
		objects[4] = originalGraph;
		return objects;
	}

	protected boolean isTransitionTau(Transition transition) {
		return transition.getLabel().startsWith("tau ") || transition.getLabel().isEmpty();
	}

	protected Set<String> getProcessAlphabet() {
		Set<String> alphabet = new HashSet<String>();

		for (Transition t : coverabilityGraph.getEdges()) {
			if (!isTransitionTau(t)) {
				alphabet.add(t.getLabel());
			}
		}

		return alphabet;
	}

	protected void addNoisyTransition(State source, State target, String activity) {
		coverabilityGraph.addTransition(source.getIdentifier(), target.getIdentifier(), activity);
		Transition transition = coverabilityGraph.findTransition(source.getIdentifier(), target.getIdentifier(), activity);
		ArrayList<Object> list = new ArrayList<Object>(3);
		list.add(transition.getSource().getIdentifier());
		list.add(transition.getTarget().getIdentifier());
		list.add(transition.getIdentifier());
		weights.add(list, costWrongFlow);
	}

	protected void cloneGraph(CoverabilityGraph originalGraph, StartStateSet originalStartIds, AcceptStateSet originalAcceptIds) {
		coverabilityGraph = new ExtendedCoverabilityGraph(originalGraph.getLabel(), costActivityNotInProcess);

		// clone the nodes
		for (State s : originalGraph.getNodes()) {
			coverabilityGraph.addState(s.getIdentifier());
			coverabilityGraph.getNode(s.getIdentifier()).setAccepting(s.isAccepting());
		}
		// clone the transitions
		for (Transition transition : originalGraph.getEdges()) {
			coverabilityGraph.addTransition(transition.getSource().getIdentifier(), transition.getTarget().getIdentifier(),
					transition.getIdentifier());
		}

		// set of starting / accepting nodes
		startIds.addAll(originalStartIds);
		acceptIds.addAll(originalAcceptIds);
	}

	protected void bypassSilent() {
		// bypass silent transitions (in this phase we add transitions -> possible violation of determininsm)
		Set<Transition> processTransitions = new HashSet<Transition>(coverabilityGraph.getEdges());
		for (Transition transition : processTransitions) {
			if (isTransitionTau(transition)) {
				State target = transition.getTarget();
				for (Transition t : coverabilityGraph.getOutEdges(target)) {
					coverabilityGraph.addTransition(transition.getSource().getIdentifier(), t.getTarget().getIdentifier(),
							t.getIdentifier());
				}
			}
		}
	}

	protected void addWeights() {
		Set<String> alphabet = getProcessAlphabet();

		for (State state : coverabilityGraph.getNodes()) {
			for (String activity : alphabet) {
				boolean activityFound = false;
				// check all outgoing transitions and verify that from the current state we can execute the activity
				for (Transition transition : coverabilityGraph.getOutEdges(state)) {
					if (transition.getLabel().equals(activity)) {
						activityFound = true;
						break;
					}
				}
				// this activity, from this state represents a deviation, we have to add it, with higher
				if (!activityFound) {
					addNoisyTransition(state, getDeviationTargetState(state, activity), activity);
				}
			}
		}
	}

	protected State getDeviationTargetState(State source, String activity) {
		State goal = null;
		// first check downward
		Pair<State, Integer> nextDownward = getDeviationTargetStateDownward(source, activity, 0, new HashSet<Object>());
		if (nextDownward != null) {
			goal = nextDownward.getFirst();
		} else {
			// if no downward state is found, then try to look for upward states
			Pair<State, Integer> nextUpward = getDeviationTargetStateUpward(source, activity, 0, new HashSet<Object>());
			if (nextUpward != null) {
				// actually, if no downward state, we hope to have found
				goal = nextUpward.getFirst();
			} else {
				// if we reach this point it means that we do not have this activity nor on the path towards the end, not from the origin.
				// it basically means that this activity belongs to a different "branch" of the process. in this case, we think it's a good
				// idea to just consider the first activity for this branch, from the process starting point.
				State startingPoint = coverabilityGraph.getNode(startIds.iterator().next());
				Pair<State, Integer> firstDownward = getDeviationTargetStateDownward(startingPoint, activity, 0, new HashSet<Object>());
				if (firstDownward != null) {
					// actually, this if should NEVER be false
					goal = firstDownward.getFirst();
				}
			}
		}
		return goal;
	}

	private Pair<State, Integer> getDeviationTargetStateDownward(State startingState, String activity, int depth, Set<Object> visitedStates) {
		if (stateIsCollectingActivity(startingState, activity)) {
			return new Pair<State, Integer>(startingState, depth);
		} else {
			visitedStates.add(startingState.getIdentifier());
			Pair<State, Integer> bestSoFar = null;
			for (Transition transition : coverabilityGraph.getOutEdges(startingState)) {
				if (!visitedStates.contains(transition.getTarget().getIdentifier())) {
					Pair<State, Integer> reached = getDeviationTargetStateDownward(transition.getTarget(), activity, depth + 1,
							visitedStates);
					if (reached != null && (bestSoFar == null || (bestSoFar.getSecond() > reached.getSecond()))) {
						bestSoFar = reached;
					}
				}
			}
			return bestSoFar;
		}
	}

	private Pair<State, Integer> getDeviationTargetStateUpward(State startingState, String activity, int depth, Set<Object> visitedStates) {
		if (stateIsCollectingActivity(startingState, activity)) {
			return new Pair<State, Integer>(startingState, depth);
		} else {
			visitedStates.add(startingState.getIdentifier());
			Pair<State, Integer> bestSoFar = null;
			for (Transition transition : coverabilityGraph.getInEdges(startingState)) {
				if (!visitedStates.contains(transition.getSource().getIdentifier())) {
					Pair<State, Integer> reached = getDeviationTargetStateUpward(transition.getSource(), activity, depth + 1,
							visitedStates);
					if (reached != null && (bestSoFar == null || (bestSoFar.getSecond() > reached.getSecond()))) {
						bestSoFar = reached;
					}
				}
			}
			return bestSoFar;
		}
	}

	private boolean stateIsCollectingActivity(State state, String activity) {
		for (Transition transition : coverabilityGraph.getInEdges(state)) {
			if (transition.getLabel().equals(activity)) {
				return true;
			}
		}
		return false;
	}
}
